# ******************************************************************************
# Class Constructors
# ******************************************************************************

#-------------------------------------------------------------------------------
# Construct new array triangle class
#-------------------------------------------------------------------------------
ArrayTriangle <- function(x, cumulative=T)
{
  # Construct and ArrayTriangle from data "x".

  #-----------------------------------------------------------------------------
  # ArrayTriangle class - copy constructor
  #-----------------------------------------------------------------------------
  if(missing(x)) {
    .data <- array(dim = c(10, 10,1)) # Default 10x10 triangle of NA's
    dimnames(.data) <- list(1:10, 1:10, 1)
  }
  else {
    #---------------------------------------------------------------------------
    # Determine type of data provided and create triangle accordingly
    #---------------------------------------------------------------------------
    # Array
    if(class(x)=="array") {
      .data <- x
      .names <- dimnames(x)
      if(is.null(.names))
        .names <- list( origin = 1:dim(x)[1],
                        dev = 1:dim(x)[2],
                        triangle = 1:dim(x)[3])
      dimnames(.data) <- .names
    }

    # Data.Frame
    if(class(x)=="data.frame")
      x <- as.matrix(x)

    # Matrix
    if(class(x)=="matrix") {
      .data <- array(x, dim = c(dim(x), 1))
      .names <- dimnames(x)
      if(is.null(.names))
        .names <- list( origin = 1:dim(x)[1],
                        dev = 1:dim(x)[2],
                        triangle = 1)
      else
        .names[["triangle"]] <- 1
      dimnames(.data) <- .names
    }
  }


  #-----------------------------------------------------------------------------
  # Assemble and return new object
  #-----------------------------------------------------------------------------
  .out <- list( data = .data,
                cumulative = cumulative
              )
  attr(.out, "class") <- c("ArrayTriangle", "Triangle", "list")
  return(.out)
}


#-------------------------------------------------------------------------------
# Convert LongTriangle to ArrayTriangle format
#-------------------------------------------------------------------------------
as.ArrayTriangle <- function(x)
{
  if(class(x)[1]=="ArrayTriangle") return(x)
  
  # Determine array names
  .names <- apply(x$data[,1:3], 2, unique)

  # Expand to include entire array, in case don't have complete data
  .id <- paste(x$data$origin, x$data$dev, x$data$triangle, sep='.')
  .grid <- expand.grid(.names)
  .grid$id <- paste(.grid$origin, .grid$dev, .grid$triangle, sep='.')
  .grid$data <- x$data$data[match(.grid$id, .id)]

  # Create data array
  .data <- array(.grid$data, dim=unlist(lapply(.names, length)), dimnames=.names)

  # Output ArrayTriangle
  .out <- list(data = .data,
               cumulative = x$cumulative
          )
  attr(.out, "class") <-  c("ArrayTriangle", "Triangle", "list")

  return(.out)
}





# ******************************************************************************
# Class Methods
# ******************************************************************************

#-------------------------------------------------------------------------------
# Print method for ArrayTriangle
#-------------------------------------------------------------------------------
print.ArrayTriangle <- function(x)
{
  print(x$data)
}

#-------------------------------------------------------------------------------
# Plot method for ArrayTriangle
#-------------------------------------------------------------------------------
plot.ArrayTriangle <- function(x, cumulative=T, ...)
{
  # Requires lattice library
  require(lattice, quietly=T)

  # Cumulative / Incremental
  if(cumulative==T) x <- makeCumulative(x)
  else x <- makeIncremental(x)

  # Convert ArrayTriangle to LongTriangle
  x <- as.LongTriangle(x, na.rm=T)

  xyplot(data ~ dev | triangle, groups=origin, type="l", data=x$data,
    key = list( lines = Rows(trellis.par.get("superpose.line"),
                  1:length(unique(x$data$origin))),
                text = list(lab = as.character(unique(x$data$origin))),
                columns = 2, title = "Origin Year"))

}

#-------------------------------------------------------------------------------
# Function to get cumulative triangles
#-------------------------------------------------------------------------------
makeCumulative.ArrayTriangle <- function(x)
{
  if(x$cumulative) return(x)

  .data <- apply(x$data, c(1,3), cumsum)
  .data <- aperm(.data, c(2,1,3))
  dimnames(.data)<- dimnames(x$data)
  
  x$data <- .data
  x$cumulative <- T
  return(x)
}


#-------------------------------------------------------------------------------
# Function to get incremental triangles
#-------------------------------------------------------------------------------
makeIncremental.ArrayTriangle <- function(x)
{
  if(x$cumulative==F) return(x)

  .incr <- function(x)
  {
    return(c(x[1], diff(x)))
  }

  x$data <- apply(x$data, c(1,3), .incr)
  x$data <- aperm(x$data, c(2,1,3))
  x$cumulative <- F
  return(x)
}

#-------------------------------------------------------------------------------
# Function to get latest claims
#-------------------------------------------------------------------------------
getLatest.ArrayTriangle <- function(x, drop=F)
{
  # Convert to LongTriangle
  x <- as.LongTriangle(x)

  # Perform calculation
  .out <- getLatest.LongTriangle(x)
  
  # Keep array in same format unless asked to drop a dimension
  .dimnames <- dimnames(.out$dev)
  dim(.out$latest) <- dim(.out$dev) <- c(dim(.out$dev)[1], 1, dim(.out$dev)[2])
  .dimnames <- list(origin=.dimnames[[1]], dev=1, triangle=.dimnames[[2]])
  dimnames(.out$latest) <- dimnames(.out$dev) <- .dimnames
  if(drop) {
    .out$latest <- drop(.out$latest)
    .out$dev <- drop(.out$dev)
  }
  
  return(.out)
}


#-------------------------------------------------------------------------------
# Internal: Function to get individual development factors
#-------------------------------------------------------------------------------
.getIndivDFs.ArrayTriangle <- function(x)
{
  # Make cumulative
  if(x$cumulative==F) x <- makeCumulative(x)

  # Perform calculations
  .nDevPrds <- dim(x$data)[2]
  .numerator <- x$data[ , c(2:.nDevPrds, .nDevPrds), , drop=F]
  .DFs <- .numerator / x$data
  .DFs[ , .nDevPrds, ] <- NA
  
  dimnames(.DFs) <- dimnames(x$data)

  return(.DFs)
}


