# ******************************************************************************
# Class Constructors
# ******************************************************************************

#-------------------------------------------------------------------------------
# Convert ArrayTriangle to LongTriangle format
#-------------------------------------------------------------------------------
.formatLong <- function(x, na.rm=F)
{
  .out <- expand.grid(dimnames(x))
  names(.out) <- c("origin", "dev", "triangle")
  .out$data <- as.vector(x)

  if(na.rm==T) .out <- .out[!is.na(.out$data), ]
  
  return(.out)
}


as.LongTriangle <- function(x, na.rm=F)
{
  if(class(x)[1]=="LongTriangle") return(x)

  .data <- .formatLong(x$data, na.rm)

  .out <- list( data = .data,
                cumulative = x$cumulative
              )
  attr(.out, "class") <-  c("LongTriangle", "Triangle", "list")

  return(.out)
}



# ******************************************************************************
# Class Methods
# ******************************************************************************

#-------------------------------------------------------------------------------
# Print method for LongTriangle
#-------------------------------------------------------------------------------
print.LongTriangle <- function(x)
{
  print(x$data)
}

#-------------------------------------------------------------------------------
# Plot method for LongTriangle
#-------------------------------------------------------------------------------
plot.LongTriangle <- function(x, cumulative=T, ...)
{
  # Requires lattice library
  require(lattice, quietly=T)

  # Cumulative / Incremental
  if(cumulative==T) x <- makeCumulative(x)
  else x <- makeIncremental(x)

  xyplot(data ~ dev | triangle, groups=origin, type="l", data=x$data,
    key = list( lines = Rows(trellis.par.get("superpose.line"),
                  1:length(unique(x$data$origin))),
                text = list(lab = as.character(unique(x$data$origin))),
                columns = 2, title = "Origin Year"))

}

#-------------------------------------------------------------------------------
# Function to get cumulative triangles
#-------------------------------------------------------------------------------
makeCumulative.LongTriangle <- function(x)
{
  # If already cumulative, return original
  if(x$cumulative) return(x)

  # Otherwise convert to ArrayTriangle, perform operation, convert back
  x <- as.ArrayTriangle(x)
  x$data <- apply(x$data, c(1,3), cumsum)
  x$data <- aperm(x$data, c(2,1,3))
  x$cumulative <- T
  x <- as.LongTriangle(x, na.rm=T)
  return(x)
}


#-------------------------------------------------------------------------------
# Function to get incremental triangles
#-------------------------------------------------------------------------------
makeIncremental.LongTriangle <- function(x)
{
  # If already incremental, return original
  if(x$cumulative==F) return(x)

  # Need function
  .incr <- function(x)
  {
    return(c(x[1], diff(x)))
  }

  # Otherwise convert to ArrayTriangle, perform operation, convert back
  x <- as.ArrayTriangle(x)
  x$data <- apply(x$data, c(1,3), .incr)
  x$data <- aperm(x$data, c(2,1,3))
  x$cumulative <- F
  x <- as.LongTriangle(x, na.rm=T)
  return(x)
}


#-------------------------------------------------------------------------------
# Function to get latest claims
#-------------------------------------------------------------------------------
getLatest.LongTriangle <- function(x)
{
  # Need function to get data corresponding to max dev period
  .func <- function(x) return(c(max(as.numeric(x$dev)),
    x$data[as.numeric(x$dev)==max(as.numeric(x$dev))]))

  # Remove NA's
  x$data <- x$data[!is.na(x$data$data), ]

  # Extract data
  .temp <- by(x$data, list(x$data$origin, x$data$triangle), .func)
  .temp <- do.call("rbind", as.list(.temp))

  # Set up results list (each member an array)
  .names <- apply(x$data[,c("origin", "triangle")], 2, unique)
  .dev <- array(.temp[,1], dim=unlist(lapply(.names, length)), dimnames=.names)
  .latest <- array(.temp[,2], dim=unlist(lapply(.names, length)), dimnames=.names)
  .out <- list(latest=.latest, dev=.dev)

  return(.out)
}
